@@ -1,10 +1,8 @@
 /*
- * Copyright (c) 1998-2000 Apple Computer, Inc. All rights reserved.
+ * Copyright (c) 1998-2003 Apple Computer, Inc. All rights reserved.
  *
  * @APPLE_LICENSE_HEADER_START@
  * 
- * Copyright (c) 1999-2003 Apple Computer, Inc.  All Rights Reserved.
- * 
  * This file contains Original Code and/or Modifications of Original Code
  * as defined in and that are subject to the Apple Public Source License
  * Version 2.0 (the 'License'). You may not use this file except in
@@ -25,6 +23,7 @@
 
 #include <IOKit/assert.h>
 #include <IOKit/IOBufferMemoryDescriptor.h>
+#include <IOKit/IODeviceTreeSupport.h>
 #include <IOKit/IOLib.h>
 #include <IOKit/storage/IOFDiskPartitionScheme.h>
 #include <libkern/OSByteOrder.h>
@@ -150,6 +149,8 @@ bool IOFDiskPartitionScheme::start(IOService * provider)
     {
         if ( partition->attach(this) )
         {
+            attachMediaObjectToDeviceTree(partition);
+
             partition->registerService();
         }
     }
@@ -161,6 +162,38 @@ bool IOFDiskPartitionScheme::start(IOService * provider)
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
+void IOFDiskPartitionScheme::stop(IOService * provider)
+{
+    //
+    // Clean up after the media objects we published before terminating.
+    //
+
+    IOMedia *    partition;
+    OSIterator * partitionIterator;
+
+    // State our assumptions.
+
+    assert(_partitions);
+
+    // Detach the media objects we previously attached to the device tree.
+
+    partitionIterator = OSCollectionIterator::withCollection(_partitions);
+
+    if ( partitionIterator )
+    {
+        while ( (partition = (IOMedia *) partitionIterator->getNextObject()) )
+        {
+            detachMediaObjectFromDeviceTree(partition);
+        }
+
+        partitionIterator->release();
+    }
+
+    super::stop(provider);
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+
 OSSet * IOFDiskPartitionScheme::scan(SInt32 * score)
 {
     //
@@ -387,7 +420,7 @@ bool IOFDiskPartitionScheme::isPartitionInvalid( fdisk_part * partition,
 
     // Determine whether the partition shares space with the partition map.
 
-    if ( partitionBase == fdiskBlock )  return true;
+    if ( partitionBase == fdiskBlock * mediaBlockSize )  return true;
 
     // Determine whether the partition starts at (or past) the end-of-media.
 
@@ -505,11 +538,64 @@ IOMedia * IOFDiskPartitionScheme::instantiateDesiredMediaObject(
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
-OSMetaClassDefineReservedUnused(IOFDiskPartitionScheme, 0);
+bool IOFDiskPartitionScheme::attachMediaObjectToDeviceTree( IOMedia * media )
+{
+    //
+    // Attach the given media object to the device tree plane.
+    //
+
+    IORegistryEntry * parent = this;
+
+    while ( (parent = parent->getParentEntry(gIOServicePlane)) )
+    {
+        if ( parent->inPlane(gIODTPlane) )
+        {
+            char         location[ 32 ];
+            const char * locationOfParent = parent->getLocation(gIODTPlane);
+            const char * nameOfParent     = parent->getName(gIODTPlane);
+
+            if ( locationOfParent == 0 )  break;
+
+            if ( OSDynamicCast(IOMedia, parent) == 0 )  break;
+
+            parent = parent->getParentEntry(gIODTPlane);
+
+            if ( parent == 0 )  break;
+
+            if ( media->attachToParent(parent, gIODTPlane) == false )  break;
+
+            strcpy(location, locationOfParent);
+            if ( strchr(location, ':') )  *(strchr(location, ':') + 1) = 0;
+            strcat(location, media->getLocation());
+            media->setLocation(location, gIODTPlane);
+            media->setName(nameOfParent, gIODTPlane);
+
+            return true;
+        }
+    }
+
+    return false;
+}
+
+OSMetaClassDefineReservedUsed(IOFDiskPartitionScheme, 0);
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
-OSMetaClassDefineReservedUnused(IOFDiskPartitionScheme, 1);
+void IOFDiskPartitionScheme::detachMediaObjectFromDeviceTree( IOMedia * media )
+{
+    //
+    // Detach the given media object from the device tree plane.
+    //
+
+    IORegistryEntry * parent;
+
+    if ( (parent = media->getParentEntry(gIODTPlane)) )
+    {
+        media->detachFromParent(parent, gIODTPlane);
+    }
+}
+
+OSMetaClassDefineReservedUsed(IOFDiskPartitionScheme, 1);
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
